/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Dict
 *
 * The Initial Developer of the Original Code is David Costanzo
 * Portions created by the Initial Developer are Copyright (C) 2003-2004
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

// Abstract:
//   This file contains the core implementation of the DICT protocol client.
//   It calls functions in connection-xpcom.js and utf8.js.
//
//   This file is used from context.xul, message-overlay.xul and 
//   dict-service-main.js, so it must only call functions that are defined
//   in those contexts.  Specifically, "alert()" is not defined when this
//   file is included from dict-service-main.js.

// put constants into a namespace so we don't pollute the global namespace
function DictState () {

    // class contants
    this.Start                = 0;
    this.WantClientAck        = 1;
    this.WantDefinition       = 2;
    this.WantDefinitionsDone  = 3;
    this.WantStrategiesInfo   = 4;
    this.WantStrategies       = 5;
    this.WantMatchesInfo      = 6;
    this.WantMatches          = 7;
    this.Done                 = 8;
    this.Error                = -1;
}

const DICTSTATE = new DictState();


function DictQueryType () {

    this.Recursive  = 0; // Define word.
                         // If no definitions, match word.
                         // If only one match, define it.

    this.MatchOnly  = 1; // Match word.

    this.DefineOnly = 2; // Define word.
}

const DICTQUERYTYPE = new DictQueryType();


function DictObserver(listener, connection, word, database, querytype) {

    this._listener   = listener;
    this._connection = connection;
    this._word       = word;       // encoded in UTF-8
    this._database   = database;

    this._bufferedData = '';
    this._state = DICTSTATE.Start;

    this._queryType = querytype;

    this._definitions  = {};

    this._createDate = new Date();
}


// Reports that there was a protocol error.
// This should only happen if the server doesn't obey the protocol
// (or if there are bugs in the client state machine)
DictObserver.prototype.reportStateMachineError = function ()
{
    this._state = DICTSTATE.Error;

    var exception = 
        'Unknown problem encountered when defining "' + this._word + '".\n' +
        'If problem presists, please file a bug report on the Dict project.\n';

    throw exception;
}

// Reports that the requested database does not exist on the server
DictObserver.prototype.reportUnknownDatabase = function ()
{
    this._state = DICTSTATE.Error;

    var exception = 
        this._connection.host + ' does not have a database called "' + this._database + '".\n' +
        'Please specify a different database.\n';

    throw exception;
}

// Reports that the server does not have any databases installed
DictObserver.prototype.reportNoDatabases = function ()
{
    this._state = DICTSTATE.Error;

    var exception = 
        this._connection.host + ' does not have any databases installed.\n' +
        'Please use a different server.\n';

    throw exception;
}

// Reports that the client received an access denied error.
DictObserver.prototype.reportAccessDenied = function ()
{
    this._state = DICTSTATE.Error;

    var exception = 
        this._connection.host + ' requires user authentication, but ' + 
        'the Dict extension does not support authentication.\n' + 
        'If this feature is important to you, please let us know by filing a bug on the Dict project.';

    throw exception;
}

// Reports that the client received an access denied error.
DictObserver.prototype.reportTemporarilyUnavailable = function ()
{
    this._state = DICTSTATE.Error;

    var exception = 
        this._connection.host + ' is temporarily unavailable, but expects to be available soon.\n' +
        'Please wait a short while and retry your request.';

    throw exception;
}


// Reports that the client received an access denied error.
DictObserver.prototype.reportServerShuttingDown = function ()
{
    this._state = DICTSTATE.Error;

    var exception = 
        this._connection.host + ' is being shut down and will not be able to define\n' + 
        'words in the near-future.\n' +
        'Please change your preferences to use a different server.';

    throw exception;
}

// Pops a DICT response line from the _bufferedInput member variable.
// On success, returns true and sets this._responseLine, this._responseCode, 
// and this._responseLineBody.
// Return false, if there wasn't a complete response line.
DictObserver.prototype.popResponseLine = function ()
{
    var responseLinePattern = /^(\d+)\s*(.*)\n/;
    
    var result = this._bufferedData.match(responseLinePattern);
    if (result == null) {
        return false;
    }

    this._responseLine     = result[0];
    this._responseCode     = result[1];
    this._responseLineBody = result[2];
    
    // remove the response line from the buffered input
    var lineEndIndex   = this._bufferedData.indexOf('\n');
    this._bufferedData = this._bufferedData.slice(lineEndIndex + 1);

    // Handle error codes that should result in immediately
    // reporting an error to the user regardless of the current state.
    if (this._responseCode == 530 ||
        this._responseCode == 531 ||
        this._responseCode == 532) {
        // access denied
        this.reportAccessDenied();
    }
    else if (this._responseCode == 420) {
        // server is temporarily unavailable
        this.reportTemporarilyUnavailable();
    }
    else if (this._responseCode == 421) {
        // server is temporarily unavailable
        this.reportServerShuttingDown();
    }
    else if (this._responseCode == 550) {
        // invalid database
        this.reportUnknownDatabase();
    }
    else if (this._responseCode == 554) {
        // no databases present
        this.reportNoDatabases();
    }

    return true;
}

// Performs a graceful shutdown
DictObserver.prototype.gracefulShutdown = function ()
{
    this._connection.sendData('QUIT\r\n');
    this._bufferedData = '';
    this._state = DICTSTATE.Done;
    
    this._connection.disconnect();
}


// Parses and returns all buffered input up to (but not including) 
// the sequence CRLF . CRLF.  The multiline input is then removed 
// from the buffered input.
// Returns null if the multiline input is not complete.
//
// The DICT protocol uses multiline input of this form to return 
// strategies, matches, and definitions. 
DictObserver.prototype.popMultilineInput = function () 
{
    // NOTE: We should like for \r\n.\r\n, but all "\r" are removed
    //       from the input stream to make regexp parsing easier.
    const sentinal = "\n.\n";

    // Check to see if we have the entire list
    var endIndex = this._bufferedData.indexOf(sentinal);
    if (endIndex == -1) {
        return null;  // the multiline input isn't complete
    }
    
    // pop off and return the multiline input
    var multilineData = this._bufferedData.slice(0, endIndex);
    
    // remove the multiline input and sentinal sequence from the buffered input
    this._bufferedData = this._bufferedData.slice(endIndex + sentinal.length);

    // According to RFC 2229, Section 2.4.3, if a line of text begins 
    // with two periods, the client must treat this as a single, 
    // literal period.
    multilineData = multilineData.replace(/^\.\./gm, '.');

    return multilineData;
}

DictObserver.prototype.onStreamDataAvailable = function (
    request, 
    inStream, 
    offset, 
    count
) 
{
    if (this._state == DICTSTATE.Done ||
        this._state == DICTSTATE.Error) {
        // sometimes if we perform a graceful shutdown, this handler
        // gets called again.  Any operation will throw an exception,
        // so it's best to just ignore the call.
        return;
    }

    try {
        // read the data from the socket into our buffer
        var newData = this._connection.readData(0, count);

        // strip out carriage returns, since they complicate regexp usage
        this._bufferedData += newData.replace(/\r/g, '');

        // do {
        //   {handle data}
        // } while( {there's enough data to handle} );
        do {
            var previousLength = this._bufferedData.length;

            var matchExp;
            var word;

            switch (this._state) {

            case DICTSTATE.Start:
                // get the 220 "hello" response that is sent when we connect

                if (this.popResponseLine()) {
                    
                    if (this._responseCode == 220) {
                        // 220 <text> <capabilities> <msg-id> CRLF
                        // (this is the session init sequence)
                        this._state = DICTSTATE.WantClientAck;
                    }
                    else {
                        this.reportStateMachineError();
                    }
                }
                break;


            case DICTSTATE.WantClientAck:
                // get the 250 ok resonse for our "CLIENT" command

                if (this.popResponseLine()) {
                    if (this._responseCode == 250) {

                        if (this._queryType == DICTQUERYTYPE.MatchOnly) {
                            // We have pipelined a MATCH request.
                            // We expect to see the matches next.
                            this._state = DICTSTATE.WantMatchesInfo;
                        }
                        else {
                            // We have pipelined a DEFINE request.
                            // We expect to see the definitions next.
                            this._state = DICTSTATE.WantDefinitionsDone;
                        }
                    }
                    else {
                        this.reportStateMachineError();
                    }
                }
                break;

            case DICTSTATE.WantDefinitionsDone:

                if (this.popResponseLine()) {
                    if (this._responseCode == 150) {
                        // 150: n definitions returned
                    }
                    else if (this._responseCode == 151) {
                        // 151: <word> <database> <name>
                        // The definition follows this line
                        this._state = DICTSTATE.WantDefinition;
                    }
                    else if (this._responseCode == 250) {

                        // We have all definitions
                        this._listener.onHaveDefinitions(
                            this._word,
                            this._definitions);

                        this.gracefulShutdown();
                    }
                    else if (this._responseCode == 550) {
                        // 550 invalid database

                        // TODO: Handle this gracefully
                        this.reportStateMachineError();
                    }
                    else if (this._responseCode == 552) {
                        // 552 no match

                        // Present the user with a nearest match 
                        
                        // The "." strategy means "use whatever is best for
                        // for interactive spell-checking", according to 
                        // RFC 2229, section 3.3.1.
                        //
                        this._connection.sendData('MATCH ' + this._database + ' . "' + this._word + '"\r\n');
                        this._state = DICTSTATE.WantMatchesInfo;
                    }
                    else {
                        this.reportStateMachineError();
                    }
                }
                break;

            case DICTSTATE.WantDefinition:
                
                // 151 <word> <database> CRLF
                // <definition line 1>   CRLF
                // <definition line 2>   CRLF
                // . . . 
                // <definition line n>   CRLF
                // . CRLF
                
                // determine if we have the entire definition
                var definitionText = this.popMultilineInput();
                if (definitionText == null) {
                    break; // wait for the rest of the definition
                }

                var definition = new Object;

                // 151 <word> <database> CRLF
                var result = this._responseLineBody.match(/^(\S+)\s+\S+\s+(.*)$/);
                if (result != null) {
                    definition.word   = result[1];
                    definition.source = result[2];
                }

                // Store the definition
                definition.text = definitionText;

                // Add this definition to the set of definitions.
                // this._definitions maps the source name to an array of 
                // definition objects.
                if (this._definitions[definition.source] != null) {
                    this._definitions[definition.source].push(definition);
                }
                else {
                    this._definitions[definition.source] = new Array ();
                    this._definitions[definition.source].push(definition);
                }

                this._state = DICTSTATE.WantDefinitionsDone;
                break;

            case DICTSTATE.WantStrategiesInfo:

                // DEAD CODE: 
                //   This state is not reachable, since the code never
                //   sends the "SHOW STRAT" command.  However,
                //   this state code is correct, so I have kept it intact.

                if (this.popResponseLine()) {
                    if (this._responseCode == 111) {
                        // 111 n stragegies available
                        this._state = DICTSTATE.WantStrategies;
                    }
                    else if (this._responseCode == 555) {
                        // 555 No strategies available
                        this.gracefulShutdown();
                        this._listener.onNoDefinitionsFound(this._word);
                    }
                    else {
                        // unexpected response
                        this.reportStateMachineError();
                    }
                }
                break;

            case DICTSTATE.WantStrategies:

                // DEAD CODE: 
                //   This state is not reachable, since the code never
                //   sends the "SHOW STRAT" command.  However,
                //   the state code is correct, so I have kept it intact.

                // <strategy 1> <description 1> CRLF
                // <strategy 2> <description 2> CRLF
                // . . . 
                // <strategy n> <description n> CRLF
                // . CRLF

                // determine if we have the entire strategy list
                var strategies = this.popMultilineInput();
                if (strategies == null) {
                    break; // wait for the rest of the strategies
                }


                // search for the "Levenshtein distance one"
                var haveLevStrategy = false;
                // <strategy> <description> CRLF
                matchExp = /^(\S+)\s+(.*)$/gm;
                while ((result = matchExp.exec(strategies)) != null) {
                    var strategy    = result[1];
                    var description = result[2];

                    if (strategy == 'lev') {
                        haveLevStrategy = true;
                    }
                }

                if (haveLevStrategy) {
                    // Use the "lev" strategy to find nearest matches
                    this._connection.sendData('MATCH ' + this._database + ' lev "' + this._word + '"\r\n');
                    this._state = DICTSTATE.WantMatchesInfo;
                }
                else {
                    // No good matching strategies
                    this.gracefulShutdown();
                    
                    this._listener.onNoDefinitionsFound(this._word);
                }

                break;

            case DICTSTATE.WantMatchesInfo:

                if (this.popResponseLine()) {
                    if (this._responseCode == 250) {
                        // 250 ok
                        // This is from the MATCH request.  Ignore it.
                    }
                    else if (this._responseCode == 152) {
                        // 152 n matches found
                        this._state = DICTSTATE.WantMatches;
                    }
                    else if (this._responseCode == 552) {
                        // 552 no match
                        this.gracefulShutdown();
                        if (this._queryType == DICTQUERYTYPE.MatchOnly) {
                            this._listener.onNoMatchesFound(this._word);
                        }
                        else {
                            this._listener.onNoDefinitionsFound(this._word);
                        }
                    }
                    else {
                        // unexpected response
                        this.reportStateMachineError();
                    }
                }
                break;

            case DICTSTATE.WantMatches:

                // <database 1> <word 1> CRLF
                // <database 2> <word 2> CRLF
                // . . . 
                // <database n> <word n> CRLF
                // . CRLF
                
                // determine if we have the entire strategy list
                var matches = this.popMultilineInput();
                if (matches == null) {
                    break; // wait for the rest of the matches
                }

                // <database> <word> CRLF
                var words    = new Object;
                matchExp = /^(\S+)\s+(.*)$/gm;
                while ((result = matchExp.exec(matches)) != null) {
                    var database    = result[1];
                    word            = result[2];

                    var normalizedWord = word.toLowerCase();

                    words[normalizedWord] = true;
                }

                // REVISIT: There has got to be a more direct way
                //   of counting the number of elements in an associative
                //   array.
                var totalMatches = 0;
                for (word in words) {
                    totalMatches++;
                }

                if (this._queryType == DICTQUERYTYPE.MatchOnly) {
                    // We are only looking for matches.
                    if (totalMatches == 0) {
                        // This is probably a bug in either the server or the
                        // client, since we should not have transitioned to
                        // the WantMatches state if there were no matches.
                        this._listener.onNoMatchesFound(this._word);
                        this.gracefulShutdown();
                    }
                    else {
                        // We found some matches.
                        this._listener.onHaveClosestMatches(this._word, words);
                        this.gracefulShutdown();
                    }
                }
                else {
                    // this is a "resurive" query.

                    if (totalMatches == 0) {
                        // There are no matches.

                        // This is probably a bug in either the server or the
                        // client, since we should not have transitioned to
                        // the WantMatches state if there were no matches.
                        this.gracefulShutdown();

                        this._listener.onNoDefinitionsFound(this._word);
                    }
                    else if (totalMatches == 1) {
                        // There is exactly one match.
                        // Rather than present just one match, we can save the
                        // user some trouble and just request the definition
                        // of the match.

                        for (word in words) {
                            // Request a definition for the only match

                            // sanitize the word
                            word = word.replace(/^"/g, '');
                            word = word.replace(/"$/g, '');
                            word = dictUcs4ToUtf8(dictUtf8ToUcs4(word));

                            this._connection.sendData('DEFINE ' + this._database + ' "' + word + '"\r\n');
                            this._word  = word;
                            this._state = DICTSTATE.WantClientAck;
                        }
                    }
                    else {
                        // There is more than one close match
                        this._listener.onHaveClosestMatches(this._word, words);

                        this.gracefulShutdown();
                    }
                }

                break;

            case DICTSTATE.Done:
                break;

            default:
                reportStateMachineError();
                break;
            }
            
        } while (previousLength != this._bufferedData.length);
    }
    catch (ex) {
        this._connection.disconnect();
        this._listener.onError(ex);
    }
}

DictObserver.prototype.onStreamClose = function (
    status
) 
{
    var closeDate = new Date();
    var delta = closeDate.getTime() - this._createDate.getTime();
    //alert('query took ' + delta + ' ms');

    if (this._state != DICTSTATE.Done &&
        this._state != DICTSTATE.Error) {
        this._listener.onError(
            'Unable to connect to "' + this._connection.host + '".\n' + 
            'If the problem persists, try using a different server.');
    }
}


function DictClient() {
}

DictClient.prototype = 
{

    // defineWord() is the main entry point to the DICT client.
    // This function contacts the DICT server and asks for the
    // definition of a given word.  It is asynchornous and the
    // definitions (or errors) are returned by callbacks on
    // the "listener" parameter.
    //
    // host - 
    //     The host running the DICT server
    //
    // port - 
    //     The port on which the DICT server is listening.
    //     This is usually 2628.
    //
    // word - 
    //     The word to define.  This should be encoded in USC-4.
    // 
    // listener - 
    //     A special "listener" object which handles the server's response.
    //     This object must define the following functions:
    //
    //       onHaveDefinitions: function (word, definitions)
    //          Called when DictClient has the definitions.
    //
    //          word - 
    //              The word that was actually defined, encoded in UTF-8.
    //              This may be different from the requested word, if
    //              the exact word was not found and only one match 
    //              existed.
    //
    //          definitions -
    //              An associated array that maps the name of the definition's source
    //              (the dictionary) to an array of definitions.
    //              Each definition is an object with one field "text",
    //              which is a UTF-8 encoded definition of the word.
    //
    //              The following code shows how to iterate over all definitions:
    //
    //                for (var source in definitions) {
    // 
    //                  var i;
    //                  for (i = 0; i < definitions[source].length; i++) {
    //
    //                    var def = definitions[source][i].text;
    //                    <do something with def>
    //                  }
    //                }
    //
    //       onHaveClosestMatches: function (word, matches)
    //          Called when DictClient has a list of close matches.
    //
    //          word    - 
    //              The word for which there are close matches.
    //              This is encoded in UTF-8.
    //
    //          matches - 
    //              An associative array whose keys are strings,
    //              each of which is a close match for a word 
    //              that has a definition.
    //              The keys are encoded in UTF-8.
    //
    //       onNoDefinitionsFound : function (word) 
    //          Called when the DictClient can not find any definitions.
    //
    //          word - 
    //              the word passed into defineWord() encoded in UTF-8.
    //
    //       onError : function (error) 
    //          Called when the DictClient encountered an unexpected error.
    //
    //          error - 
    //              A plain-text explaination of what went wrong.
    //              This string is encoded in USC-4.
    //
    // database - [optional]
    //    The database to define the word.
    //    If this parameter is not given, '*' will be used.
    //
    defineWord : function(host, port, word, listener, database)
    {

        // check to see if we are in an offline mode
        var ioService = Components.classes["@mozilla.org/network/io-service;1"].getService(Components.interfaces.nsIIOService);

        if (ioService.offline) {
            listener.onError(
                'You cannot define words while offline.\n' +
                'To go online, uncheck "Work Offline" from the File menu.');
            return;
        }


        // if no database was specified, default to '*' (all databases)
        if (database == null) {
            database = '*';
        }

        var connection = new DictCBSConnection();

        connection.connect(host, port, (void 0), true, null);

        // construct a UTF-8 representation of word
        var utf8Word = dictUcs4ToUtf8(word);

        var observer = new DictObserver(
            listener,
            connection,
            utf8Word,
            database,
            DICTQUERYTYPE.Recursive);

        // To save time, pipe-line the CLIENT and DEFINE commands,
        // rather than wait for the "220 ok" handshake.
        // The server is required to support command pipe-lining.
        // This makes a measurable speed gain when there is high latency.
        connection.sendData('CLIENT Mozilla Dict extension\r\n');
        connection.sendData('DEFINE ' + database + ' "' + utf8Word + '"\r\n');
        connection.startAsyncRead(observer);
    },

    // matchWord() is a secondary entry point to the DICT client.
    // This function contacts the DICT server and asks for the
    // closest matches of a given word.  It is asynchornous and the
    // matches (or errors) are returned by callbacks on the "listener"
    // parameter.
    //
    // host - 
    //     The host running the DICT server
    //
    // port - 
    //     The port on which the DICT server is listening.
    //     This is usually 2628.
    //
    // word - 
    //     The word to find matches for.  This should be encoded in USC-4.
    // 
    // listener - 
    //     A special "listener" object which handles the server's response.
    //     This object must define the following functions:
    //
    //       onHaveClosestMatches: function (word, matches)
    //          Called when DictClient has a list of close matches.
    //
    //          word    - 
    //              The word for which there are close matches.
    //              This is encoded in UTF-8.
    //
    //          matches - 
    //              An associative array whose keys are strings,
    //              each of which is a close match for a word 
    //              that has a definition.
    //              The keys are encoded in UTF-8.
    //
    //       onNoMatchesFound : function (word) 
    //          Called when the DictClient can not find any matches.
    //
    //          word - 
    //              the word passed into defineWord() encoded in UTF-8.
    //
    //       onError : function (error) 
    //          Called when the DictClient encountered an unexpected error.
    //
    //          error - 
    //              A plain-text explaination of what went wrong.
    //              This string is encoded in USC-4.
    //
    // database - [optional]
    //    The database to define the word.
    //    If this parameter is not given, all databases on host will be
    //    searched.
    //
    // strategy - [optional]
    //    The strategy to use when matching.
    //    If this parameter is not given, a strategy that works well
    //    for catching simple misspellings will be used.
    //
    matchWord : function(host, port, word, listener, database, strategy)
    {

        // check to see if we are in an offline mode
        var ioService = Components.classes["@mozilla.org/network/io-service;1"].getService(Components.interfaces.nsIIOService);

        if (ioService.offline) {
            listener.onError(
                'You cannot define words while offline.\n' +
                'To go online, uncheck "Work Offline" from the File menu.');
            return;
        }


        // if no database was specified, default to '*' (all databases)
        if (database == null) {
            database = '*';
        }

        // if no strategy was specified, default to '.'
        if (strategy == null) {
            strategy = '.';
        }

        var connection = new DictCBSConnection();

        connection.connect(host, port, (void 0), true, null);

        // construct a UTF-8 representation of word
        var utf8Word = dictUcs4ToUtf8(word);

        var observer = new DictObserver(
            listener,
            connection,
            utf8Word,
            database,
            DICTQUERYTYPE.MatchOnly);

        // To save time, pipe-line the CLIENT and MATCH commands,
        // rather than wait for the "220 ok" handshake.
        // The server is required to support command pipe-lining.
        // This makes a measurable speed gain when there is high latency.
        connection.sendData('CLIENT Mozilla Dict extension\r\n');
        connection.sendData('MATCH ' + database + ' ' + strategy + ' "' + utf8Word + '"\r\n');

        connection.startAsyncRead(observer);
    }
}

